% 'exoShape" is a derived software by Stanley Durrleman, Copyright (C) INRIA (Asclepios team), All Rights Reserved, 2006-2009, version 1.0
%--------------------------------------------------------------------------
% Based on MATCHINE v1.0 software.
% Copyright Universit Paris Descartes
% Contributor: Joan Alexis GLAUNES (2006)
% alexis.glaunes@mi.parisdescartes.fr
% 
% This software is a computer program whose purpose is to calculate an optimal 
% diffeomorphic transformation in 3D-space that allows to match two datasets 
% like points, curves or surfaces.
% 
% This software is governed by the CeCILL-B license under French law and
% abiding by the rules of distribution of free software. You can use, 
% modify and/ or redistribute the software under the terms of the CeCILL-B 
% license as circulated by CEA, CNRS and INRIA at the following URL
% "http://www.cecill.info". 
% 
% As a counterpart to the access to the source code and rights to copy,
% modify and redistribute granted by the license, users are provided only
% with a limited warranty  and the software's author, the holder of the
% economic rights, and the successive licensors have only limited 
% liability. 
% 
% In this respect, the user's attention is drawn to the risks associated
% with loading, using, modifying and/or developing or reproducing the
% software by the user in light of its specific status of free software,
% that may mean that it is complicated to manipulate, and that also
% therefore means that it is reserved for developers and experienced
% professionals having in-depth computer knowledge. Users are therefore
% encouraged to load and test the software's suitability as regards their
% requirements in conditions enabling the security of their systems and/or 
% data to be ensured and, more generally, to use and operate it in the 
% same conditions as regards security. 
% 
% The fact that you are presently reading this means that you have had
% knowledge of the CeCILL-B license and that you accept its terms.
%--------------------------------------------------------------------------
 
% Any use of this code should make reference to:
% - S. Durrleman, X. Pennec, A. Trouve and N. Ayache, Statistical Models of Sets of Curves and Surfaces based on Currents, Medical Image Analysis, (2009), DOI: 10.1016/j.media.2009.07.007
% 
% AND to at least one of the following references:
% - [for the use of 'landmark or measure matching']:
% J. Glaunes, L. Younes and A. Trouve, Diffeomorphic matching of distributions: A new approach for unlabelled point-sets and sub-manifolds matching, Proc. of the 2004 IEEE Computer Society Conference on Computer Vision and Pattern Recognition. (CVPR'04), Vol. 2, pp. 712--718, DOI: 10.1109/CVPR.2004.81 
% 
% - [for the use of 'curve matching']:
% J. Glaunes, A. Qiu, M. Miller, L. Younes, Large Deformation Diffeomorphic Metric Curve Mapping, International Journal of Computer vision, Springer, 2008, Vol. 80, No. 3, pp. 317--336, DOI: 10.1007/s11263-008-0141-9 
% 
% - [for the use of 'surface matching']:
% M. Vaillant and J. Glaunes, Surface Matching via Currents, Proc. of Information Processing in Medical Imaging (IPMI'05), Lecture Notes in Computer Science vol. 3565, Springer 2005, pp. 381--392

function [fft3k_d fft3k k] = noyau3D_PAIR(grille,lambda)
  % ne fonctionne que pour des dimensions paires
  longX = grille.long(1);
  longY = grille.long(2);
  longZ = grille.long(3);
  pas = grille.pas;
  k = zeros(grille.long);
  demilongX = longX/2;
  demilongY = longY/2;
  demilongZ = longZ/2;
  lambda2 = lambda*lambda;
  for i = 0:demilongX
   for j = 0:demilongY
    for n = 0:demilongZ
     k(i+1,j+1,n+1) = exp(-pas^2*(i^2+j^2+n^2)/lambda2);
    end
   end
  end
  
  vx0 = 1:(demilongX+1);
  vxm = longX:-1:(demilongX+2);
  vx2 = 2:demilongX;

  vy0 = 1:(demilongY+1);
  vym = longY:-1:(demilongY+2);
  vy2 = 2:demilongY;

  vz0 = 1:(demilongZ+1);
  vzm = longZ:-1:(demilongZ+2);
  vz2 = 2:demilongZ;

  k(vxm,vy0,vz0) = k(vx2,vy0,vz0);
  k(vx0,vym,vz0) = k(vx0,vy2,vz0);
  k(vx0,vy0,vzm) = k(vx0,vy0,vz2);
  k(vxm,vym,vz0) = k(vx2,vy2,vz0);
  k(vxm,vy0,vzm) = k(vx2,vy0,vz2);
  k(vx0,vym,vzm) = k(vx0,vy2,vz2);
  k(vxm,vym,vzm) = k(vx2,vy2,vz2);
  
  fft3k = fftn(k); % transformee de Fourier du noyau
  fft3k = real(fft3k); % pour eviter les imprecisions numeriques
  fft3k_d = fft3k(1:(demilongX+1),:,:); % matrice decimee pour tenir compte des symmetries de fft3k
end
